#include <mh_log.h>
#include "AStar.h"

namespace clcs {

std::vector<int16_t> AStar::startSearch()
{
	// create root node and add it to N and Q
	shared_ptr<Node> root = std::make_shared<Node>();
	root->pL = std::vector<int16_t>(inst->m, 1);
	root->parent = nullptr;
	addNode(root);

	try {
		// run the main loop
		while (!Q.empty()) {
			shared_ptr<Node> v = Q.top().node;
			Q.pop();
			// check if node is outdated (required because priority_queue in c++ doesn't support removal of arbitrary elements)
			auto N_known = N.find(v->pL);
			if (N_known != N.end()) {
				bool outdated = false;
				auto it_known = N_known->second.begin();
				while (it_known != N_known->second.end() && !outdated && it_known->first >= v->u_v) {
					if ((it_known->second > v->l_v) || (it_known->second >= v->l_v && it_known->first > v->u_v)) {
						outdated = true;
					}
					++it_known;
				}
				if (outdated) {
					continue;
				}
			}
//			mh::out() << "Expanding "; printNode(v, false);
			// find non-dominated sigma
			map<int, vector<int16_t>> sigma_nd = sol->findSigmaNd(*v);
			// check if we arrived at a complete node (i.e., an optimal solution was found)
			if (sigma_nd.empty()) {
				return sol->deriveSolution(v.get());
			}
			// create an expansion for every letter from sigma_nd
			++stat_nodes_expanded;
			for (auto &it : sigma_nd) {
				int letter = it.first;
				vector<int16_t> &pL = it.second;

				shared_ptr<Node> v_ext = sol->expandNode(v, letter, pL);
				auto N_rel = N.find(pL);

				if (N_rel == N.end()) {
					addNode(v_ext); // it's the first node for this sub-problem, so just add it to N and Q
				} else {
					// compare with existing nodes of same sub-problem and possibly merge them
					auto it_rel = N_rel->second.begin();
					bool insert = true;
					while (it_rel != N_rel->second.end()) {
						int16_t u_v_rel = it_rel->first;
						int16_t l_v_rel = it_rel->second;
						if (insert && v_ext->u_v > u_v_rel) {
						    insert = false;
						    addNode(v_ext, &N_rel->second, &it_rel);  // insert node in order
						}
						if (l_v_rel >= v_ext->l_v && u_v_rel >= v_ext->u_v) {
						    insert = false;
						    ++stat_nodes_ignored;
						    break; // the extension is inferior and can be ignored
						}
						if (v_ext->l_v >= l_v_rel && v_ext->u_v >= u_v_rel) {
						    it_rel = N[v_ext->pL].erase(it_rel);
						    ++stat_nodes_merged;
						} else {
						    ++it_rel; // next element
						}
					}
					if (insert) {
						addNode(v_ext, &N_rel->second, &it_rel); // insert node at the end of list
					}
				}
			}
		}
        // when memory is exceeded, catch the proven UB value on the problem
	} catch (std::bad_alloc &ex) {
		mh::out() << "Out of memory!" << std::endl;
		mh::out() << "UB: " << Q.top().ub << std::endl;
	}
	return std::vector<int16_t>(); // there is no solution
}

void AStar::printStatistics()
{
	mh::out() << "Node Stats (created, expanded, ignored, merged): ";
	mh::out() << stat_nodes_created << ", " << stat_nodes_expanded << ", ";
	mh::out() << stat_nodes_ignored << ", " << stat_nodes_merged << std::endl;
	mh::out() << "Sizes (N, Q): " << N.size() << ", " << Q.size() << std::endl;
	mh::out() << "Max Nested (N_rel): " << stat_max_nested << std::endl;
}

/**  @v added into corresp. N_rel; 
     if N_rel is empty -> create a linked list and add node @v
     or update corresp. entry in N_rel  at given position @pos
 */
void AStar::addNode(const shared_ptr<Node> &v, std::list<std::pair<int16_t, int16_t>> *list,
					std::list<std::pair<int16_t, int16_t>>::iterator *pos)
{
        // mh::out() << " creating "; printNode(v);
	Q.push(QNode(v, ub(*v, inst)));
	++stat_nodes_created;
	if (list == nullptr) {
	    N[v->pL] = std::list<std::pair<int16_t, int16_t>>({{v->u_v, v->l_v}});
	} else {
		list->emplace(*pos, v->u_v, v->l_v);
		if (stat_max_nested < list->size()) {
		    ++stat_max_nested;
		}
	}
}
/** print partial solution that corresp. to node @v into an .out file */
void AStar::printNode(const shared_ptr<Node> &v, bool withParent)
{
	if (v->l_v > 0 && withParent) {
		mh::out() << "from ((";
		for (int i = 0; i < inst->m; i++) {
		     mh::out() << v->parent->pL[i];
		     if (i != inst->m - 1 ) {
			 mh::out() << ",";
		     }
		}
		mh::out() << "), " << v->parent->l_v << ", ";
		mh::out() << v->parent->u_v;
		mh::out() << ") with '";
		mh::out() << inst->int2Char.at(inst->S[0][v->pL[0]-2]) << "' -> ";
	}

	mh::out() << "Node ((";
	for (int i = 0; i < inst->m; i++) {
		mh::out() << v->pL[i];
		if (i != inst->m - 1 ) {
			mh::out() << ",";
		}
	}
	mh::out() << "), ";
	mh::out() << v->l_v << ", ";
	mh::out() << v->u_v << ")";
	if (!withParent) {
		mh::out() << std::endl;
	}
}

}
