#include <memory>
#include <random>
#include "BeamSearch.h"

namespace clcs {

/** mhlib external parameters */
mh::int_param bw("bw", "beam width parameter", 10, 0, 100000);
mh::int_param kbest("kbest", "(restricted) filtering parameter", 0, 0, 100000);
mh::int_param kext("kext", "pre-restriction parameter", 100000, 0, 100000);
mh::int_param guidance("guidance", "0: UB; 1: H_prob; 2: H_pow; 3: EX; 4: Piece-by-piece EX, 5: Min-Avg", 0, 0, 5);
mh::int_param filtering("filtering", "turn on Filter procedure", 1, 0, 1);

void BeamSearch::startSearch()
{
	// create root node
	shared_ptr<Node> root = std::make_shared<Node>();
	root->pL = std::vector<int16_t>(inst->m, 1);
	root->parent = nullptr;
	s_best = root;
	// initialize beam and add root node to it
	auto beam = vector<shared_ptr<Node>>();
	beam.push_back(root);

	// run beam search until beam is empty
	while (!beam.empty())
	{
		vector<NodeExtension> extensions = extendAndEvaluate(beam);
		//cout << "|Vext| " << extensions.size() << endl;
		if (pruning() == 1)
		    prune(extensions);
		if (filtering() == 1)
		    filter(extensions);
		beam.clear();
		for(int i = 0; i < (int) extensions.size() && (int) beam.size() < k_bw; i++) {
		    beam.push_back(extensions[i].node);
		}
	}
}
/** Extend all nodes from the set of given nodes @beam (priority values of each node are pre-calculated) */
vector<NodeExtension> BeamSearch::extendAndEvaluate(const vector<shared_ptr<Node>>& beam)
{
	auto extensions = vector<NodeExtension>();

	for (const shared_ptr<Node> &v : beam) {
	     map<int, vector<int16_t>> sigma_nd = sol->findSigmaNd(*v);
	     for (auto &it : sigma_nd) {
			int letter = it.first;
			vector<int16_t> pL = it.second;
			// create new child node by appending letter a and calculate greedy value
			NodeExtension extension = NodeExtension();
			shared_ptr<Node> newChild = sol->expandNode(v, letter, pL);
			extension.node = newChild;
			extension.h = clcs_eval::greedy(*extension.node, inst);
			extensions.push_back(extension);

			// check if we found a new valid best solution
			if (newChild->l_v > s_best->l_v && newChild->u_v == inst->p) {
				s_best = newChild;
			}
	     }
	}
	// take best solutions according to the best-next heuristic
	std::sort(extensions.begin(), extensions.end());

	if ((int) extensions.size() > max_extensions) {
		extensions.resize(max_extensions);
	}

	// apply main heuristic and sort extensions
	int px = 0;
	if (guidance() == (int) Heuristic::H_PROB) { // calculate p-value if the H_prob heuristic used
	     px = determine_p(extensions);  
	}
	for (NodeExtension &extension : extensions) {
	     if (guidance() > 0) {
			extension.h = h(*extension.node, inst, px);
			if (h2 != nullptr) {
				extension.h2 = h2(*extension.node, inst, px);
			}
		} else {  // use UB as a guidance:
			extension.h = ub(*extension.node, inst);
		}
	}
	std::sort(extensions.begin(), extensions.end(), std::greater<NodeExtension>());

	return extensions;
}

/** determine p-value in H_prob heuristic **/
int BeamSearch::determine_p( vector<NodeExtension> &extensions )
{
	int p_min = 100000;
	for (auto &ext : extensions) {
		Node &v = *ext.node.get();

		int h_v = inst->P.size() - v.u_v;
		if (h_v < p_min)
			p_min = h_v;
	}

	int fl_term = 100000;
	for (NodeExtension &ext : extensions) {
		Node &v = *ext.node.get();
		// min len:
		int l_min = 100000;
		for (int i = 0; i < inst->m; ++i) {
			if (l_min > (int) inst->S[i].size() - v.pL[i] + 1 - p_min)
				l_min = inst->S[i].size() - v.pL[i] + 1 - p_min;
		}
		int fl_term_v = std::floor((l_min) / (inst->sigma + 0.0));
		if (fl_term_v < fl_term)
			fl_term = fl_term_v;
	}

	if (fl_term == 0) /* p_min + fl_term == 0 */
		return 1;
	else
		return p_min + fl_term; /*p_min + fl_term */
}

/** Prune function in GBS framework */
void BeamSearch::prune(vector<NodeExtension> &extensions)
{        
	extensions.erase(std::remove_if(extensions.begin(), extensions.end(),
			[this](const NodeExtension &e) { return ub(*e.node, inst) <= s_best->l_v; }), extensions.end());
}

bool BeamSearch::dominated_member(int index, vector<NodeExtension>& C, int k_fi) // return true if node is dominated by some nodes from k_best...
{
	int kf = (k_fi > 0) ? k_fi : kbest();
	Node *node = C[index].node.get();

	for (int i = 0; i < std::min(kf, (int) C.size()); i++)
	{
		if (i != index and isDominated(node, C)) {   // C[i] - from k_best dominates <node>
			return true;
		}
	}
	return false;
}

/** pairwise Filter procedure in GBS framework; set up acc. to parameter kbest()  */
void BeamSearch::filter(vector<NodeExtension>& extensions)   // if k_filter() > C.size(): full filtering (for all elements) applied
{
	for (int i = 0; i < (int) extensions.size(); i++)
	{
		if (dominated_member(i, extensions, kbest())) // it is dominated w.r.t. some of best k_best nodes
		{
		    extensions.erase(extensions.begin() + i); // free the address
		    i--;
		}
	}
}

vector<int16_t> BeamSearch::getSolution()
{
	return sol->deriveSolution(s_best.get());
}

bool BeamSearch::isDominated(Node* v,  vector<NodeExtension>& extensions)  
{
	for (int i = 0; i < (int)extensions.size() && i < k_best; i++)
	{
		Node &v2 = *extensions[i].node.get(); //address of the content
		if (v != &v2 && v->u_v <= v2.u_v) {
		    if (sol->check_domination(v->pL, v2.pL)) {
			return true;
		    }
		}
	}
	return false;
}
/** Randomized Greedy heuristic executed using corresp. @seed and corresp. random value @p_fill (parameter of the procedure) */
void BeamSearch::startGreedy(double p_fill, unsigned int seed)
{
	// initialize seed and random number generation (used only for randomized greedy)
	bool is_randomized = p_fill > 0;
	if (seed == 0) {
	    // no seed specified -> generate random seed
	    std::random_device rd;
	    seed = rd();
	}
	std::mt19937 rand_gen(seed);
	std::uniform_real_distribution<> dis(0.0, 1.0);

	// initialize root node
	shared_ptr<Node> v = std::make_shared<Node>();
	v->pL = std::vector<int16_t>(inst->m, 1);
	v->parent = nullptr;
	s_best = v;

	// start actual greedy search
	map<int, vector<int16_t>> sigma = sol->findSigmaNd(*v);
	while(!sigma.empty()) {

		bool skipHeuristic = false;
		if (is_randomized && v->u_v < inst->p) {
			int letterForMatch = inst->P[v->u_v];
			if (dis(rand_gen) < p_fill && sigma.count(letterForMatch) > 0) {
			    // expand with letter that forms a strong matching
			    v = sol->expandNode(v, letterForMatch, sigma[letterForMatch]);
			    skipHeuristic = true;
			} else if (sigma.size() > 1) {
				// only consider the remaining letters
				sigma.erase(letterForMatch);
			}
		}

		if (!skipHeuristic) { // skip this step in case we already took the letter forming a strong match
			double bestVal = INFINITY;
			shared_ptr<Node> bestChild = nullptr;
			for (auto &it : sigma) {
				int letter = it.first;
				vector<int16_t> pL = it.second;
				shared_ptr<Node> child = sol->expandNode(v, letter, pL);
				double g = clcs_eval::greedy(*child, inst);
				if (g < bestVal) {
					bestVal = g;
					bestChild = child;
				}
			}
			v = bestChild;
		}

		sigma = sol->findSigmaNd(*v);
	}
	s_best = v;
}

}
