#include "CLCS_evaluation.h"

using namespace clcs;

/** MinAvg heuristic: parameter to break ties (Eucledean 2-distance norm serves as default) */
mh::int_param k_term_pow("k_term_pow", "k for k-term of avg-min heuristic", 2, 1, 100);

namespace clcs_helper {

/** EX calculation 
    decompose the exponent value due to stability reasons when exponent is extremely high)
*/
long double stable_pow(double v, int sigma, double power, int step)
{
	long double val = v;
	while (power > step) {
	       val = pow(val, pow(sigma, step));
	       power -= step;
	}
	val = pow(val, pow(sigma, power));
	return 1.0 - val;
}
/* Approximate l: l-t term of the sum; Sigma: alphabet size; pi: value of 1-E[Y_l] */
long double Approx_term_k(double l, int Sigma, long double pi)
{
	long double term = 1.0;
	if (pi < 1E-250)
		return 0.0;

	long double pxInterm = log(pi) + l * log(Sigma);
	if (pxInterm >= 300)
	    term = 1.0;
	else { // in-between...
	    long double px = exp(pxInterm);
	    long double tildeP = px * (-1.0 - pi / 2.0);
	    if (abs(tildeP) < 1E-15)// 1-e^x ~ -x
		return -tildeP;
	    term = exp(tildeP);
	}
	return (long double) 1.0 - term;
}

int l_min(const Node &v, const CLCS_inst *inst)
{
	int l_min = INT_MAX;
	for (int i = 0; i < inst->m; i++) {
	     if (l_min > (int) inst->S[i].size() - v.pL[i] + 1)
		 l_min = (int) inst->S[i].size() - v.pL[i] + 1;
	}
	return l_min;
}

/** term_k: calcaute k-th product (1-x)^y  **/
long double term_k(const Node &v, const CLCS_inst *inst, int k)
{
	long double pi = 1.0;
	for (int i = 0; i < inst->m; ++i)
		pi *= (inst->P_m[k][((int) inst->S[i].size() - v.pL[i] + 1)]);// * inst->P_m[ inst->p -  v.u_v ][ k ] );
	pi *= inst->P_m[inst->p - v.u_v][k]; // calculate x

	if (pi >= 1.0E-10) {
		pi = ((long double) 1.0) - pi;
		pi = stable_pow(pi, inst->sigma, k, 20);
	} else
		pi = Approx_term_k(k, inst->sigma, pi);
	return pi;
}

long double fastEX(const Node &v, const CLCS_inst *inst, int bp, int ep)
{
	if (bp > ep or bp < 0)
		return 0.0;
	else if (bp == ep)
		return term_k(v, inst, bp);
	else {
		int mid = floor((bp + ep) / 2.0);
		long double mid_val = term_k(v, inst, mid);
		if (mid_val <= 1E-6)
			return (ep - mid + 1) * mid_val + fastEX(v, inst, bp, mid - 1);
		if (mid_val >= 1.0 - 1E-6)
			return mid - bp + 1 + fastEX(v, inst, mid + 1, ep);
		else
			return mid_val + fastEX(v, inst, bp, mid - 1) + fastEX(v, inst, mid + 1, ep);
	}
}

}


namespace clcs_eval {

double greedy(const Node &v, const CLCS_inst *inst)
{
	Node *parent = v.parent.get();
	double g = 0.0;
	for (int i = 0; i < inst->m; i++) {
	     g += (v.pL[i] - parent->pL[i] + 1) / (double) (inst->S[i].size() - parent->pL[i] + 1);
	}
	/*if (parent->u_v < inst->p && v.u_v == parent->u_v) {
		g += 1.0 / (inst->p - parent->u_v);
	}*/
         if (parent->u_v < inst->p) {   
             int q = (v.u_v != parent->u_v) ? 1 : 0; // strong match	
             g += 1.0 / (inst->p - parent->u_v + q);
	}
	return g;
}


int ub1(const Node &v, const CLCS_inst *inst)
{
	int ub1 = 0;
	vector<int> sigma_s;
	vector<int> min_occurances_node((int) inst->sigma, (int) inst->S[0].size());

	for (int a = 0; a < inst->sigma; a++) {
		for (int i = 0; i < inst->m; i++) {
			if (v.pL[i] - 1 < (int16_t) inst->S[i].size()) //not feasible letter
			{
				if (inst->occurance_positions[i][a][v.pL[i] - 1] < min_occurances_node[a])
					min_occurances_node[a] = inst->occurance_positions[i][a][v.pL[i] - 1];
			} else min_occurances_node[a] = 0; // some of indices of pL exceed max indices...
		}
	}
	for (unsigned int i = 0; i < min_occurances_node.size(); i++) {
		//if(min_occurances_node[i] >= 1)
		ub1 += min_occurances_node[i]; // sum of the c_a
	}
	return v.l_v + ub1;
}

int ub2(const Node &v, const CLCS_inst *inst)
{
	int ub2 = inst->S[0].size();
	for (int i = 0; i < inst->m - 1; ++i) {
		if (clcs::ub() == 1) {   // CLCS-based Ubs
			if (ub2 > inst->M[i][v.pL[i] - 1][v.pL[i + 1] - 1][v.u_v])
				ub2 = inst->M[i][v.pL[i] - 1][v.pL[i + 1] - 1][v.u_v];
		} else {// lcs scoring matrices...

			if (ub2 > inst->M_lcs[i][v.pL[i] - 1][v.pL[i + 1] - 1])
				ub2 = inst->M_lcs[i][v.pL[i] - 1][v.pL[i + 1] - 1];
		}
	}
	return v.l_v + ub2;
}

/* LCS upper bound, min (UB_1, UB_2) */
int ub(const Node &v, const CLCS_inst *inst)
{
	int ub_1 = clcs_eval::ub1(v, inst);
	int ub_2 = clcs_eval::ub2(v, inst);

	return std::min(ub_1, ub_2);
}


long double h_prob(const Node &v, const CLCS_inst *inst, int p)
{
	long double h_val = 1.0;
	for (int i = 0; i < inst->m; ++i) {
		h_val *= inst->P_m[p][inst->S[i].size() - v.pL[i] + 1];
	}
	// check this?
	/*if ( (int)inst->P.size() - v.u_v  <= p)
		 h_val *= inst->P_m[ inst->P.size() - v.u_v ][ p ]; */

	return h_val;
}

long double h_power(const Node &v, const CLCS_inst *inst, int p = 0)
{
	long double h_pow = 1.0;
	double a = 1.8, b = 0.1, c = 0.1;
	// TODO: use actual parameters for a, b, c
	double px = (a * exp(-b * inst->m) + c);

	long double q_min = INT_MAX;
	for (int i = 0; i < inst->m; i++) {
		long double q_i = (long double) (inst->S[i].size() - v.pL[i] + 1) / (inst->p - v.u_v + 1);
		h_pow *= q_i;
		if (q_min > q_i)
			q_min = q_i;
	}
	h_pow = pow(h_pow, px);
	return h_pow * q_min;
}

long double h_min_avg(const Node &v, const CLCS_inst *inst, int p = 0)
{
	int l_min = clcs_helper::l_min(v, inst);
	return (long double) l_min / (inst->p - v.u_v + 1);
}

long double k_norm(const Node &v, const CLCS_inst *inst, int p = 0)
{
	long double k_norm = 0;
	for (int i = 0; i < inst->m; i++) {
		k_norm += pow(((long double)(inst->S[i].size() - v.pL[i] + 1) / (inst->p - v.u_v + 1)), k_term_pow());
	}
	return k_norm;
}


long double ex(const Node &v, const CLCS_inst *inst, int p)
{
	int l_min = clcs_helper::l_min(v, inst);
	long double ex = clcs_helper::fastEX(v, inst, inst->p - v.u_v, l_min); // l_m change for UB...
	return ex;
}

}
