#include <fstream>
#include <iostream>
#include <string>
#include <math.h>
#include <assert.h>
#include <map>
#include <climits>
#include <algorithm>
#include "mh_log.h"
#include "mh_util.h"
#include "CLCS_inst.h"

using namespace std;
using namespace mh;

namespace clcs {
/** mhlib external parameters */
mh::int_param ub("ub", "upper bound used as a guidance: 0: LCS based UB: 1: CLCS UB (min of two, using the CLCS DP-based matrices)", 0, 0, 1); 
mh::int_param pruning("pruning", "turn on the Prune method", 1, 0, 1);

CLCS_inst::CLCS_inst(const std::string &fname) : filename(fname)
{
	cout << "Read instance " << fname << endl;
	ifstream is(fname);
	if (!is)
		mherror("Cannot open the problem instance file", fname);

	is >> m >> sigma >> p;
	if (m < 1)
		mherror("Invalid number of sequences when reading problem instance", tostring(m));
	if (sigma < 2)
		mherror("Invalid alphabet size when reading problem instance", tostring(sigma));

	int count = 0;
	int len = 0;
	n = 0;
	string sq;
	// read constrained string P
	is >> len >> sq;
	P.reserve(len);
	for (int j = 0; j < len; ++j) { // read its characters
		int a;
		if (char2Int.count(sq[j]) == 0) {

		    char2Int[sq[j]] = count;
		    int2Char.push_back(sq[j]);
		    a = count;
		    count++;
		} else
		    a = char2Int[sq[j]];
		this->P.push_back(a); // store into constrained string P
	}
	p = P.size();

	for (int i = 0; i < m; i++) {
		// read S_i as string, preceded by the length and transform it in vector<int>
		len = 0;
		string seq;
		is >> len >> seq;
		if (!is || len < 1)
		    mherror("Instance file does not contain enough data");
		if (len != int(seq.size()))
		    mherror("Stated sequence length does not correspond to string length in problem instance", tostring(i));
		vector<int> seqs(len, 0);
		S.push_back(seqs);
		for (int j = 0; j < len; ++j) { // read characters...
			int a;
			if (char2Int.count(seq[j]) == 0) {

				char2Int[seq[j]] = count;
				int2Char.push_back(seq[j]);
				a = count;
				count++;
			} else
				a = char2Int[seq[j]];
			S[i][j] = a; // coding the letter with number...
		}
		n = max(n, len);
	}

	structure_occurances(); // create occurrences structure...
	structure_embeddings(); // create embedding structures...
	store_M();   // create M structure...
	P_matrix(); // populate heuristic (probabilistic) matrix...
	UB = 0;
	vector<int> minc(sigma, INT_MAX); // minimal numbers of appearances of each letter in all sequences
	for (int i = 0; i < m; i++) {
		vector<int> c(sigma, 0);  // number of appearances of each letter in S_i
		for (int a : S[i])
		     c[a]++;
		for (int a = 0; a < sigma; a++)
		     minc[a] = min(c[a], minc[a]);
		c.clear();
	}
	for (int i = 0; i < sigma; i++) {
	     if (minc[i] >= 1)
	         UB += minc[i];
	}
}

std::vector<int> CLCS_inst::occurances_string_letter(std::vector<int> &str, int a)
{
	//occurances of letter <a> in string str: 0 - means that it does not appear: vector[pl] denoted the number of occurrences of letter a starting from s[pl, end];
	vector<int> occurances_letter(n, 0);
	// predecessor of S_i for  letter <letter> :
	int number = 0;
	for (int i = n - 1; i >= 0; i--) // successors of X
	{
	     if (str[i] == a)
		 number++;
	     occurances_letter[i] = number;
	}
	return occurances_letter;
}

std::vector<int> CLCS_inst::successor_string_letter(std::vector<int> &str, int a)
{
	vector<int> successor_letter(n + 1, 0);
	int size = (int) str.size();
	int number = size + 1; // starting position meaning that no successor of letter <a>
	for (int i = size - 1; i >= 0; --i) {
		if (str[i] == a) {
			number = i + 1;
			successor_letter[i] = number;
		} else
			successor_letter[i] = number;
	}
	if (str[0] == a)
		successor_letter[0] = 1;

	return successor_letter;
}

std::vector<std::vector<int>> CLCS_inst::occurances_all_letters(vector<int> &str)
{
	vector<vector<int>> occurances_all(sigma, vector<int>(n, 0));
	vector<int> occurances_letter(n, 0);

	for (int a = 0; a < this->sigma; a++) // iterate through each letter from alphabet and find Occ vector for it for each input string <str>
		occurances_all[a] = occurances_string_letter(str, a);

	return occurances_all;
}

vector<vector<int>> CLCS_inst::succesors_all_letters(vector<int> &str)
{
	vector<vector<int>> successors_all(sigma, vector<int>(n + 1, 0));
	for (int a = 0; a < this->sigma; a++) // iterate for all letters from alphabet
		successors_all[a] = successor_string_letter(str, a);
	return successors_all;

}

void CLCS_inst::structure_occurances()
{       // procedure to establish preprocessing structure Occs
	occurance_positions = vector<vector<vector<int> > >(m, vector<vector<int>>(sigma, vector<int>(n, 0)));
	successors = vector<vector<vector<int> > >(m, vector<vector<int>>(sigma, vector<int>(n + 1, 0)));
	for (int i = 0; i < this->m; i++) {
		vector<int> s_i = S[i];
		occurance_positions[i] = occurances_all_letters(S[i]);
		successors[i] = succesors_all_letters(S[i]);
	}
}


void CLCS_inst:: P_matrix()
{
        // P-values from Mousavi and Tabataba (2012): required to establis EX guidance, and here is preprocessed
	P_m = std::vector<std::vector<double>>(n + 1, std::vector<double>(n + 1 , 0.0)); // reserve memory
	for (int q = 0; q <= n; ++q)
        {
	     for (int k  = 0; k <= n; ++k)
	     {
		  if (k == 0)
	              P_m[k][q] = 1;
		  else
		  if(k > q)
		      P_m[k][q] = 0;
		  else
	              P_m[k][q] = ((double) 1.0 / sigma ) * P_m[k-1][q - 1] + ( (double) sigma - 1.0)/ (sigma) * P_m[k][q-1];
	     }
        }
}

vector<vector<int>> CLCS_inst::lcs_m_ij(int i, int j)   // creating LCS M_ij score matrix (of dimension |s_i| x |s_j|)
{
	vector<vector<int>> m_ij(S[i].size() + 1, vector<int>(S[j].size() + 1, 0));

	for (int x = (int) S[i].size(); x >= 0; x--) {
		for (int y = (int) S[j].size(); y >= 0; y--) {
			if (x == (int) S[i].size())
				m_ij[x][y] = 0;
			else if (y == (int) S[j].size())
				m_ij[x][y] = 0;
			else if (S[i][x] == S[j][y]) // x + 1 and y + 1
				m_ij[x][y] = m_ij[x + 1][y + 1] + 1;
			else
				m_ij[x][y] = std::max(m_ij[x][y + 1], m_ij[x + 1][y]);   // corresponds to m_ij(i,j) = LCS(s_i(x), s_j(y))
		}
	}
	return m_ij;
}

vector<vector<vector<int16_t>>> CLCS_inst::clcs_m_ij(int i, int j)   // creating CLCS M_ij score matrix (of dimension |s_i| x |s_j| x |P|)
{
	vector<vector<vector<int16_t>>> m_ij(S[i].size() + 1,
			vector<vector<int16_t>>(S[j].size() + 1, vector<int16_t>(p + 1, INT16_MIN)));

	// initializing borders
	for (int x = 0; x <= (int) S[i].size(); x++) {
		m_ij[x][S[j].size()][p] = 0;
	}
	for (int y = 0; y <= (int) S[j].size(); y++) {
		m_ij[S[j].size()][y][p] = 0;
	}
	// calculate values of scoring matrix (from bottom right corner to upper left corner)
	for (int x = (int) S[i].size() - 1; x >= 0; x--) {
		for (int y = (int) S[j].size() - 1; y >= 0; y--) {
			for (int z = p; z >= 0; z--) {
			     if (S[i][x] == S[j][y]) {
					if (z != p && S[i][x] == P[z]) {
					    // strong match
					    m_ij[x][y][z] = 1 + m_ij[x + 1][y + 1][z + 1];
					} else {
					    // weak match
					    m_ij[x][y][z] = 1 + m_ij[x + 1][y + 1][z];
					}
			     } else {
					// no match
					m_ij[x][y][z] = std::max(m_ij[x + 1][y][z], m_ij[x][y + 1][z]);
		          }
		       }
		}
	}
	return m_ij;
}

void CLCS_inst::store_M() //<< creating M structure...
{
	if (ub() != 0) {
		mh::out() << "Using CLCS scoring matrices..." << std::endl;
	}

	for (int i = 0; i < m - 1; i++) {
		if (ub() == 0)// basic UB for the LCS problem
		    M_lcs.push_back(lcs_m_ij(i, i + 1));
		else // CLCS DP-based scoring matrices
		    M.push_back(clcs_m_ij(i, i + 1));
	}
}

void CLCS_inst::write(ostream &ostr, int detailed) const
{
	ostr << "m=" << m << " n=" << n << " |Sigma|=" << sigma
		 << " UB=" << UB << endl;
	if (detailed > 0) {
		// write all sequences
		for (int i = 0; i < m; i++) {
			ostr << "S" << i << "=";
			bool first = true;
			for (int a : S[i]) {
			     if (first)
				  first = false;
				else
				  ostr << ",";
			     ostr << a;
			}
			ostr << endl;
		}
		ostr << "Sigma mapping: ";
		for (int i = 0; i < sigma; i++)
		     ostr << i << "->" << int2Char[i] << " ";
		ostr << endl;
		ostr << "Constrained string P:" << P.size() << endl;
		for (int px : this->P)
		     ostr << px << " ";
		ostr << endl;
	}
}

void CLCS_inst::structure_embeddings()
{       
        // preporcess Embed data structure
	embed_end = std::vector<std::vector<int>>(m, std::vector<int>());
	for (int i = 0; i < m; i++) {
		int p_u = p - 1;
		std::vector<int> s_i = S[i];
		for (unsigned int j = s_i.size() - 1; j >= 0 && p_u >= 0; j--) {
			if (s_i[j] == P[p_u]) {
				embed_end[i].push_back(j);
				p_u--;
			}
		}
		std::reverse(embed_end[i].begin(), embed_end[i].end());
		embed_end[i].push_back(s_i.size()); // border case for once the constraint is fulfilled
	}
}

} // namespace lcps
