#include <string>
#include <map>
#include <vector>
#include <assert.h> 

#ifndef CLCS_INST_H
#define CLCS_INST_H

namespace clcs {

/** A class representing a Constrained Longest Common Subsequence (CLCS) problem instance. */
extern mh::int_param alg;
/** which upper bound to use **/
extern mh::int_param ub; 
/** prune method **/
extern mh::int_param pruning;

class CLCS_inst {
public:
	/** Constructor creates an instance by reading it from the specified file. */
	CLCS_inst(const std::string &fname);

public:
	const std::string filename;    ///< Name of instance file.
	int m;                   ///< Number of input strings.
	int n;                   ///< Length of the longest input string.
	int p;                         ///< Length of the constrained string
	int sigma;               ///< Size of the alphabet.
	int UB;                   ///< An (trivial) upper bound for the solution length.
	std::vector<std::vector<int>> S; ///< Vector of input sequences encoded by integers from [0,sigma).
	std::vector<std::vector<std::vector<int>>> occurance_positions;
	///< structure for reading the number of occurrences of letter <a> of some string i<=m, starting from pl to the end of string (s_i[pos, end]): call asoccurences[i][a][pos]
	std::vector<std::vector<std::vector<int>>> successors;  ///< Succ structure (preprocessing)
	std::vector<std::vector<std::vector<int>>> M_lcs; ///< M stores score matrices (relaxed version only M_{i, i+1}) (LCS case)
	std::vector<std::vector<std::vector<std::vector<int16_t>>>> M; ///< M stores score matrices (relaxed version only M_{i, i+1}) (CLCS case)
	std::vector<char> int2Char; ///< Translation table for internal letters corresponding to integers in [0,sigma) to real alphabet.
	std::map<char, int> char2Int;///< Translation table for real alphabet into integers in [0,sigma).
	std::vector<int> P; ///< Constraint string.
	std::vector<std::vector<int>> embed_end;  ///< Embed structure (preprocessing)
	std::vector<std::vector<double>> P_m;
	/** Method for creating a constant problem instance object by reading the instance data from
	 the given file. */
	static const CLCS_inst *createFromFile(const std::string &fname) {
		return new CLCS_inst(fname);
	}

	/** Write instance data in a human readable format. */
	void write(std::ostream &os, int detailed = 0) const;

	/** Destructor. */
	~CLCS_inst() {
	}

private:
	/** preprocessing Occurance-based structure 
            that is required for an efficient UB_1 calculation  */
	std::vector<int> occurances_string_letter(std::vector<int> &str, int letter);
	std::vector<std::vector<int>> occurances_all_letters(std::vector<int> &str);

	/** Succ data structure  */
	std::vector<std::vector<int>> succesors_all_letters(std::vector<int> &str);
	std::vector<int> successor_string_letter(std::vector<int> &str, int a);

	/** UB_2: preprocessing structures for efficient calcualtion of upper bound UB_2  */
	std::vector<std::vector<int>> lcs_m_ij(int i, int j); // creating lcs version of m_ij (score) matrix ...
	std::vector<std::vector<std::vector<int16_t>>> clcs_m_ij(int i, int j); // creating clcs version of M_ij (score) matrix ...
	void store_M();
	/** populate Occ structure */
	void structure_occurances();
	/** populate Embed structure */
	void structure_embeddings();
	/** populate P-matrix for deriving H-heuristic (m-CLCS problem; for BS search used as search guidance) */
	void P_matrix();
};

}

#endif /* CLCS_INST_H */
